from .reader import read_varint, read_binary_uint8, read_binary_int32
from .writer import write_varint, write_binary_uint8, write_binary_int32


class BlockInfo(object):
    is_overflows = False
    bucket_num = -1

    def write(self, buf):
        # Set of pairs (`FIELD_NUM`, value) in binary form. Then 0.
        write_varint(1, buf)
        write_binary_uint8(self.is_overflows, buf)

        write_varint(2, buf)
        write_binary_int32(self.bucket_num, buf)

        write_varint(0, buf)

    def read(self, buf):
        while True:
            field_num = read_varint(buf)
            if not field_num:
                break

            if field_num == 1:
                self.is_overflows = bool(read_binary_uint8(buf))

            elif field_num == 2:
                self.bucket_num = read_binary_int32(buf)


class Block(object):
    dict_row_types = (dict, )
    tuple_row_types = (list, tuple)
    supported_row_types = dict_row_types + tuple_row_types

    def __init__(self, columns_with_types=None, data=None, info=None,
                 types_check=False, received_from_server=False):
        self.columns_with_types = columns_with_types or []
        self.data = data or []
        self.types_check = types_check

        if data and not received_from_server:
            # Guessing about whole data format by first row.
            first_row = data[0]

            if self.types_check:
                self.check_row_type(first_row)

            if isinstance(first_row, dict):
                self.dicts_to_rows(data)
            else:
                self.check_rows(data)

        self.info = info or BlockInfo()

        super(Block, self).__init__()

    def dicts_to_rows(self, data):
        column_names = [x[0] for x in self.columns_with_types]

        check_row_type = False
        if self.types_check:
            check_row_type = self.check_dict_row_type

        for i, row in enumerate(data):
            if check_row_type:
                check_row_type(row)

            self.data[i] = [row[name] for name in column_names]

    def check_rows(self, data):
        expected_row_len = len(self.columns_with_types)

        got = len(data[0])
        if expected_row_len != got:
            msg = 'Expected {} columns, got {}'.format(expected_row_len, got)
            raise ValueError(msg)

        if self.types_check:
            check_row_type = self.check_tuple_row_type
            for row in data:
                check_row_type(row)

    def get_columns(self):
        return self.data

    def get_rows(self):
        if not self.data:
            return self.data

        # Transpose results: columns -> rows.
        n_columns = self.columns
        n_rows = self.rows

        flat_data = [None] * n_columns * n_rows

        for j in range(n_columns):
            column = self.data[j]

            for i in range(n_rows):
                flat_data[i * n_columns + j] = column[i]

        # Make rows from slices.
        rv = [None] * n_rows
        for i in range(n_rows):
            offset = i * n_columns
            rv[i] = tuple(flat_data[offset:offset + n_columns])

        return rv

    def check_row_type(self, row):
        if not isinstance(row, self.supported_row_types):
            raise TypeError(
                'Unsupported row type: {}. dict, list or tuple is expected.'
                .format(type(row))
            )

    def check_tuple_row_type(self, row):
        if not isinstance(row, self.tuple_row_types):
            raise TypeError(
                'Unsupported row type: {}. list or tuple is expected.'
                .format(type(row))
            )

    def check_dict_row_type(self, row):
        if not isinstance(row, self.dict_row_types):
            raise TypeError(
                'Unsupported row type: {}. dict is expected.'
                .format(type(row))
            )

    @property
    def columns(self):
        return len(self.data)

    @property
    def rows(self):
        return len(self.data[0]) if self.columns else 0




from .. import errors
from ..reader import read_binary_str, read_binary_str_fixed_len
from ..writer import write_binary_str, write_binary_str_fixed_len
from ..util import compat
from .base import CustomItemColumn


class String(CustomItemColumn):
    ch_type = 'String'
    py_types = compat.string_types

    # TODO: pass user encoding here

    def read(self, buf):
        return read_binary_str(buf)

    def _read_null(self, buf):
        self.read(buf)

    def write(self, value, buf):
        write_binary_str(value, buf)

    def _write_null(self, buf):
        self.write('', buf)


class FixedString(String):
    ch_type = 'FixedString'

    def __init__(self, length, **kwargs):
        self.length = length
        super(FixedString, self).__init__(**kwargs)

    def read(self, buf):
        return read_binary_str_fixed_len(buf, self.length).strip('\x00')

    def write(self, value, buf):
        try:
            write_binary_str_fixed_len(value, buf, self.length)
        except ValueError:
            raise errors.TooLargeStringSize()




from struct import Struct

from ..util import compat
from .base import Column
from .intcolumn import UInt64Column

if compat.PY3:
    from queue import Queue
else:
    from Queue import Queue


class ArrayColumn(Column):
    """
    Nested arrays written in flatten form after information about their
    sizes (offsets really).
    One element of array of arrays can be represented as tree:
    (0 depth)          [[3, 4], [5, 6]]
                      |               |
    (1 depth)      [3, 4]           [5, 6]
                   |    |           |    |
    (leaf)        3     4          5     6

    Offsets (sizes) written in breadth-first search order. In example above
    following sequence of offset will be written: 4 -> 2 -> 4
    1) size of whole array: 4
    2) size of array 1 in depth=1: 2
    3) size of array 2 plus size of all array before in depth=1: 2 + 2 = 4

    After sizes info comes flatten data: 3 -> 4 -> 5 -> 6
    """
    py_types = (list, tuple)
    size_struct = Struct('<Q')

    def __init__(self, nested_column, **kwargs):
        self.size_column = UInt64Column()
        self.nested_column = nested_column
        self._write_depth_0_size = True
        super(ArrayColumn, self).__init__(**kwargs)

    def size_pack(self, value):
        return self.size_struct.pack(value)

    def size_unpack(self, buf):
        return self.size_struct.unpack(buf.read(self.size_struct.size))[0]

    def write_data(self, data, buf):
        # Column of Array(T) is stored in "compact" format and passed to server
        # wrapped into another Array without size of wrapper array.
        self.nested_column = ArrayColumn(self.nested_column)
        self.nested_column.nullable = self.nullable
        self.nullable = False
        self._write_depth_0_size = False
        self._write(data, buf)

    def read_data(self, rows, buf):
        self.nested_column = ArrayColumn(self.nested_column)
        self.nested_column.nullable = self.nullable
        self.nullable = False
        return self._read(rows, buf)

    def _write_sizes(self, value, buf):
        q = Queue()
        q.put((self, value, 0))

        cur_depth = 0
        offset = 0
        nulls_map = []

        while not q.empty():
            column, value, depth = q.get_nowait()

            if cur_depth != depth:
                cur_depth = depth
                offset = 0
                if column.nullable:
                    self._write_nulls_map(nulls_map, buf)

                nulls_map = []

            if column.nullable:
                value = value or []

            offset += len(value)
            if (cur_depth == 0 and self._write_depth_0_size) or cur_depth > 0:
                buf.write(self.size_pack(offset))

            nested_column = column.nested_column
            if isinstance(nested_column, ArrayColumn):
                for x in value:
                    q.put((nested_column, x, cur_depth + 1))
                    nulls_map.append(None if x is None else False)

    def _write_data(self, value, buf):
        if self.nullable:
            value = value or []

        if isinstance(self.nested_column, ArrayColumn):
            for x in value:
                self.nested_column._write_data(x, buf)
        else:
            self.nested_column._write_data(value, buf)

    def _write_nulls_data(self, value, buf):
        if self.nullable:
            value = value or []

        if isinstance(self.nested_column, ArrayColumn):
            for x in value:
                self.nested_column._write_nulls_data(x, buf)
        else:
            if self.nested_column.nullable:
                self.nested_column._write_nulls_map(value, buf)

    def _write(self, value, buf):
        self._write_sizes(value, buf)
        self._write_nulls_data(value, buf)
        self._write_data(value, buf)

    def _read(self, size, buf):
        q = Queue()
        q.put((self, size, 0))

        data = []
        slices_series = []

        cur_depth = 0
        prev_offset = 0
        slices = []

        if self.nested_column.nullable:
            nulls_map = self._read_nulls_map(size, buf)
        else:
            nulls_map = [0] * size

        # Read and store info about slices.
        while not q.empty():
            column, size, depth = q.get_nowait()

            nested_column = column.nested_column

            if cur_depth != depth:
                cur_depth = depth

                slices_series.append((slices, nulls_map))

                if nested_column.nullable:
                    nulls_map = self._read_nulls_map(prev_offset, buf)
                else:
                    nulls_map = [0] * prev_offset

                prev_offset = 0
                slices = []

            if isinstance(nested_column, ArrayColumn):
                for _i in range(size):
                    offset = self.size_unpack(buf)
                    q.put((nested_column, offset - prev_offset, cur_depth + 1))
                    slices.append((prev_offset, offset))
                    prev_offset = offset

            # Read data
            else:
                data.extend(
                    nested_column._read_data(
                        size, buf,
                        nulls_map=nulls_map[prev_offset:prev_offset + size]
                    )
                )
                prev_offset += size

        # Build nested tuple structure.
        for slices, nulls_map in reversed(slices_series):
            nested_data = []
            for (slice_from, slice_to), is_null in zip(slices, nulls_map):
                nested_data.append(
                    None if is_null else tuple(data[slice_from:slice_to])
                )

            data = nested_data

        return tuple(data)


def create_array_column(spec, column_by_spec_getter):
    inner = spec[6:-1]
    return ArrayColumn(column_by_spec_getter(inner))



from datetime import timedelta, tzinfo


class tzutc(tzinfo):
    """
    Simple class for obtaining naive datetime without timezone.
    It prevents using whole pytz/dateutil in requirements.
    """

    ZERO = timedelta(0)

    def utcoffset(self, dt):
        return tzutc.ZERO

    def dst(self, dt):
        return tzutc.ZERO

    def tzname(self, dt):
        return 'UTC'



from calendar import timegm
from datetime import datetime

from ..util.tzinfo import tzutc
from .base import FormatColumn


class DateTimeColumn(FormatColumn):
    ch_type = 'DateTime'
    py_types = (datetime, )
    format = 'I'

    utc = tzutc()

    def after_read_item(self, value):
        return datetime.fromtimestamp(value, tz=self.utc).replace(tzinfo=None)

    def before_write_item(self, value):
        return int(timegm(value.timetuple()))



from uuid import UUID

from .base import FormatColumn
from .. import errors
from ..util import compat
from ..writer import MAX_UINT64


class UUIDColumn(FormatColumn):
    ch_type = 'UUID'
    py_types = compat.string_types + (UUID, )
    format = 'Q'

    # UUID is stored by two uint64 numbers.
    def write_items(self, items, buf):
        n_items = len(items)

        uint_64_pairs = [None] * 2 * n_items
        for i, x in enumerate(items):
            i2 = 2 * i
            uint_64_pairs[i2] = (x >> 64) & MAX_UINT64
            uint_64_pairs[i2 + 1] = x & MAX_UINT64

        s = self.make_struct(2 * n_items)
        buf.write(s.pack(*uint_64_pairs))

    def read_items(self, n_items, buf):
        s = self.make_struct(2 * n_items)
        items = s.unpack(buf.read(s.size))

        uint_128_items = [None] * n_items
        for i in range(n_items):
            i2 = 2 * i
            uint_128_items[i] = (items[i2] << 64) + items[i2 + 1]

        return uint_128_items

    def after_read_item(self, value):
        return UUID(int=value)

    def before_write_item(self, value):
        try:
            if not isinstance(value, UUID):
                value = UUID(value)

        except ValueError:
            raise errors.CannotParseUuidError(
                "Cannot parse uuid '{}'".format(value)
            )

        return value.int



from calendar import timegm
from datetime import date

from .base import FormatColumn


class DateColumn(FormatColumn):
    ch_type = 'Date'
    py_types = (date, )
    format = 'H'

    offset = 24 * 3600

    def before_write_item(self, value):
        return timegm(value.timetuple()) // self.offset

    def after_read_item(self, value):
        return date.fromtimestamp(value * self.offset)



import importlib

from .. import errors
from ..protocol import CompressionMethodByte


def get_compressor_cls(alg):
    try:
        module = importlib.import_module('.' + alg, __name__)
        return module.Compressor

    except ImportError:
        raise errors.UnknownCompressionMethod(
            "Unknown compression method: '{}'".format(alg)
        )


def get_decompressor_cls(method_type):
    if method_type == CompressionMethodByte.LZ4:
        module = importlib.import_module('.lz4', __name__)

    elif method_type == CompressionMethodByte.ZSTD:
        module = importlib.import_module('.zstd', __name__)

    else:
        raise errors.UnknownCompressionMethod()

    return module.Decompressor



from __future__ import absolute_import
from io import BytesIO

import zstd

from .base import BaseCompressor, BaseDecompressor
from ..protocol import CompressionMethod, CompressionMethodByte
from ..reader import read_binary_uint32
from ..writer import write_binary_uint32, write_binary_uint8


class Compressor(BaseCompressor):
    method = CompressionMethod.ZSTD
    method_byte = CompressionMethodByte.ZSTD

    def get_compressed_data(self, extra_header_size):
        rv = BytesIO()

        data = self.data.getvalue()
        compressed = zstd.compress(data)

        header_size = extra_header_size + 4 + 4  # sizes

        write_binary_uint32(header_size + len(compressed), rv)
        write_binary_uint32(len(data), rv)
        rv.write(compressed)

        return rv.getvalue()


class Decompressor(BaseDecompressor):
    method = CompressionMethod.ZSTD
    method_byte = CompressionMethodByte.ZSTD

    def get_decompressed_data(self, method_byte, compressed_hash,
                              extra_header_size):
        size_with_header = read_binary_uint32(self.stream)
        compressed_size = size_with_header - extra_header_size - 4

        compressed = BytesIO(self.stream.read(compressed_size))

        block_check = BytesIO()
        write_binary_uint8(method_byte, block_check)
        write_binary_uint32(size_with_header, block_check)
        block_check.write(compressed.getvalue())

        self.check_hash(block_check.getvalue(), compressed_hash)

        compressed = compressed.read(compressed_size - 4)

        return zstd.decompress(compressed)
